
// Class DynamicDataSet

// This class defines datasets whose values can be calculated.
//  e.g.
//     Outstanding Shares <=> "Net Profits" divided by "EPS"

// This class extends the DataSet class as follows:

// The constructor takes a Formula for use in calculating values

// The semantics of the storage methods are the same as in DataSet and store
// values explicitly.

// The semantics of retrieval is that if an explicit value is stored, then
// that will be used.  If no explicit value is available, then the formula is
// used to calculate the result value.

// Usage:
//    The constructors of DynamicDataSet are private methods and not
//    accessible to other classes.  Instead, there is a public class method
//    'define(name, formula-definition)' to be used.
//    Example Usage:
//        DynamicDataSet.define("PTP Margin", "100 * [PreTax Inc] / Sales");

// Also, DynamicDataSets are not defined on a per-company basis, but
// instead are defined generally.  Retrievals of a DDS for a company will
// instantiate that DDS within that company the first time a retrieval is
// attempted.

package sptossg ;

import java.util.*;

public class DynamicDataSet {

    Formula formula;
    String name;

    static Hashtable allGenericSets = new Hashtable();

    // Constructor for a DynamicDataSet - Note that this is private;
    // all definitions of DynamicDataSets should go through the public
    // static method 'define'.
    private DynamicDataSet (String name, Formula formula) {
	this.name = name;
	this.formula = formula;
	allGenericSets.put(name, this);
    }

    static public DynamicDataSet define(String name, String def) {

	char c1 = name.charAt(0);
	
	if ((c1 >= 'a' && c1 <= 'z') || (c1 >= 'A' && c1 <= 'Z')) {
	    Formula formula = Formula.parse(def);
	    if (formula == null) { return null; }
	    return new DynamicDataSet(name, formula);
	} else {
	    System.out.println("Invalid dynamic attribute name '" + name + "'. ");
	    return null;
	}
    }

    // static method to find a named DynamicDataSet
    static DynamicDataSet find(String name) {
	Object set = allGenericSets.get(name);
	if (set == null) { return null; }
	return (DynamicDataSet) set;
    }

    // External retrieve routine -
    //  Static method called with dynamic attribute name and period
    //  finds the right DynamicDataSet and then evaluates that
    //  DynamicDataSet for that period.
    static public void dump(CompanyStorage company) {
	System.out.println("Calculated Values: ");

	Enumeration e;
	e = allGenericSets.keys();
	while (e.hasMoreElements()) {
	    String name = (String) e.nextElement();
	    dump(company, name);
	}
    }

    static public void dump(CompanyStorage company, String name) {
	DynamicDataSet dds = (DynamicDataSet) allGenericSets.get(name);
	if (dds != null) {
	    dds.dumpVals(company, name);
	} else {
	    Debug.debug("No Dynamic dataset found for " + name);
	}
    }

    void dumpVals (CompanyStorage company, String name) {

	String indent = "    ";
	System.out.println(indent + name + ": Dynamic DataSet");
	indent += indent;

	Vector keys;

	// Check if it's a single variabled Dynamic DataSet
	Period p = new Period();
	String val = retrieve(company, p);
	if (val != null) {
	    keys = new Vector();
	    keys.addElement(p);
	} else {
	    keys = formula.getKeys(company);
	    if (keys == null) {
		System.out.println(indent + "No keys found. ");
		return;
	    }
	}

	Enumeration e = keys.elements();
	while (e.hasMoreElements()) {
	    p = (Period) e.nextElement();

	    val = retrieve(company, p);

	    // if we have a value for this key
	    if (val != null) {
		double nval = (new Double(val)).doubleValue();
		// round to two digits
		int nval2 = (int) (nval * 100);
		val = "" + (nval2 / 100.0);

		System.out.println(indent + p.toString() + ": " + val);
	    }
	}
    }

    public String toString() {
	return ("def([" + name + "]) = \"" + formula + "\"");
    }
	
    static public String retrieve(CompanyStorage company, String name, Period p) {
	Object set = allGenericSets.get(name);
	if (set == null) { return null; }
	DynamicDataSet dds = (DynamicDataSet) set;
	return dds.retrieve(company, p);
    }

    // doing_evaluation is used for automatic cycle detection during
    // evaluation.  For example, if A = B / 10 and B = A * 10, an
    // evaluation of either A or B will cause a undesired cycle.
    boolean doing_evaluation = false;

    //  use formula to get value
    String retrieve(CompanyStorage company, Period p) {
	double val;
	String result;

	// if we're already in an evaluation of this dynamic attribute,
	// then terminate the cycle with no value.  Note that this
	// cycle detection should allow us to define DDS formulas based
	// on one another and do retrievals as long as one of the
	// attributes has an explicit value stored.
	if (doing_evaluation) {
	    System.out.println("Warning: Evaluation cycle detected on " + name);
	    val = Double.NaN;
	} else {
	    try {
		doing_evaluation = true;
		val = formula.evaluate(company, p);
	    } finally {
		doing_evaluation = false;
	    }
	}
	if (Double.isNaN(val)) { result = null; }
	else { result = ("" + val); }
	return result;
    }

    // getKeys - Return all keys (the set of periods that might be valid
    // for this dynamic attribute), in sorted order
    Vector getKeys(CompanyStorage company) {
	return formula.getKeys(company);
    }

    // static getKeys - find the DynamicDataSet named ATTR, then get all
    // keys for it.
    static Vector getKeys(CompanyStorage company, String attr) {
	DynamicDataSet dds = find(attr);
	if (dds == null) { return null; }
	return dds.getKeys(company);
    }
}
